/*
 * systest.c -- Test code for nooks system calls
 *
 * Copyright (C) 2002 Mike Swift
 *		 2003 Christophe Augier under Mike Swift
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

#include "nooks-agent.h"

driver_t drivers[NOOKS_MAX_DRIVERS];

int driver_count = 0;
char *nook_name = NULL;
char *new_nook_name = NULL;

char * line = NULL,
	* file = NULL;

static char *nooks_stripo(char *fname)
{
	char *pt = fname;
	char *ptv;
	int len;

	if ((pt = strrchr(fname, '/')) != NULL)
		fname = pt + 1;
	len = strlen(fname);
	if (len > 2 && !strcmp(fname + len - 2, ".o"))
		len -= 2;
	else if (len > 4 && !strcmp(fname + len - 4, ".mod"))
		len -= 4;
#ifdef CONFIG_USE_ZLIB
	else if (len > 5 && !strcmp(fname + len - 5, ".o.gz"))
		len -= 5;
#endif
	else
		return fname;

	ptv = malloc(len + 1);
	if (ptv != NULL) {
		strncpy(ptv, fname, len);
		ptv[len] = '\0';
	}
	return ptv;		/* memleak */
}

OPT_LIST *nooks_recovery_options(char *module)
{
	char *modname;
	int i;

	if (*module == '/')
		modname = module;
	else
		modname = nooks_stripo(module);

	for (i = 0; recover_list && recover_list[i].name; ++i) {
		if (strcmp(recover_list[i].name, modname) == 0)
			return recover_list + i;
	}
	return NULL;
}

#define NUM_RECOVERY_DRIVERS (sizeof(recover_functions) / sizeof(driver_recovery_t))

int unload_module(char *name, int major)
{
	int status;
	int pid;

	printf("module unload called!\n");

	if ((pid = fork()) != 0) {
		waitpid(pid, &status, 0);
		writelog(" Modprobe -r %s returns %d\n", name, status);
	} else {
		char *argv[4], *envp[3];

		argv[0] = "modprobe";
		argv[1] = "-r";
		argv[2] = name;
		argv[3] = NULL;
		envp[0] = "HOME=/";
		envp[1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
		envp[2] = NULL;
		writelog(" Executing %s %s %s %s\n", argv[0], argv[1], argv[2], argv[3]);
		execve("/sbin/modprobe.nooks", argv, envp);
		
	}

	return 0;
}


int load_module(char *name, int major)
{
	int status;
	int pid;
	char new_name[20];

	printf("module load called!\n");
	// Now reload the module
	//
	if ((pid = fork()) != 0) {
		waitpid(pid, &status, 0);
		//writelog(" Modprobe %s returned %d\n", name, status);
	} else {
		char *argv[5], *envp[3];


		argv[0] = "modprobe";
		argv[1] = "-C";
		argv[2] = "/etc/modules.nooks";

		// strcpy(new_name, nook_name);
		// new_name[strlen(new_name)-1]++;

		//argv[0] = "insmod.old";
		//argv[1] = "-N";
		//argv[2] = new_name;
		//argv[0] = new_name;
		//argv[1] = "-C";
	       	//argv[2] = "/etc/modules.nooks";
		//argv[3] = name;

		argv[3] = name;
		argv[4] = NULL;

		envp[0] = "HOME=/";
		envp[1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
		envp[2] = NULL;

		writelog(" Executing %s %s %s %s\n", argv[0], argv[1], argv[2], argv[3]);
		execve("/sbin/modprobe.nooks", argv, envp);

		// writelog("Executing load_ide nook=%s\n", new_name);
		// execve("/sbin/insmod.old", argv, envp);

	}
	
	
  	return 0;
}


int cleanup_module(char *name, int major)
{

	OPT_LIST *optlist;
	int status;
	optlist = nooks_recovery_options(name);

	if (optlist != NULL) {
		GLOB_LIST *g;
		int i;

		g = optlist->opts;
		for (i = 0; i < g->pathc; i++) {
			int pid;
			//
			// call the cleanup script
			//
			if ((pid = fork()) != 0) {
				waitpid(pid, &status, 0);
				writelog(" Script %s returned %d\n",
					 g->pathv[i], status);
			} else {
				char *argv[3], *envp[3];
				argv[0] = g->pathv[i];
				argv[1] = name;
				argv[2] = NULL;
				envp[0] = "HOME=/";
				envp[1] =
				    "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
				envp[2] = NULL;
				writelog(" Executing %s %s\n", argv[0], argv[1]);
				execve(g->pathv[i], argv, envp);
			}
		}

	}
	return 0;
}

int start_network(char *name, int major)
{
	int status;
	int pid;
	//
	// Rebind the network interface
	//
	
	if ((pid = fork()) != 0) {
		waitpid(pid, &status, 0);
		writelog(" Network start script returned %d\n", status);
	} else {
		char *argv[4], *envp[3];
		argv[0] = "/sbin/ifup";
		argv[1] = name;
		argv[2] = NULL;
		envp[0] = "HOME=/";
		envp[1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
		envp[2] = NULL;
		writelog(" Executing %s %s\n", argv[0], argv[1]);
		execve("/sbin/ifup", argv, envp);
	}
	
	return (0);
}


int stop_network(char *name, int major)
{
	int status;
	int pid;
	//
	// Rebind the network interface
	//
	if ((pid = fork()) != 0) {
		waitpid(pid, &status, 0);
		writelog(" Network stop script returned %d\n", status);
	} else {
		char *argv[4], *envp[3];
		argv[0] = "/sbin/ifdown";
		argv[1] = name;
		argv[2] = NULL;
		envp[0] = "HOME=/";
		envp[1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
		envp[2] = NULL;
		writelog(" Executing %s %s\n", argv[0], argv[1]);
		execve("/etc/init.d/network", argv, envp);
	}
	return (0);
}

int unmount_fs(char *name, int major)
{
	char *fs_type = NULL;
	char *fs_dir = NULL;
	char *fs_dev = NULL;
	int flags = major;
	int result = 0;

	//
	// We need to do unmount as part of unload
	//
	char *temp = strdup(name);

	printf("unmount called!\n");
	
	fs_dir = strtok(temp, ";");
	fs_dev = strtok(NULL, ";");
	fs_type = strtok(NULL, ";");

	printf("Checkpoint 2\n");
	writelog(" Unmounting type %s, dir = %s, dev = %s, flags = %d\n",
		 fs_type, fs_dir, fs_dev, flags);

	printf("Checkpoint 3\n");
	result = umount(fs_dir);
	writelog(" umount %s returned %d\n", fs_dir, result);
	printf("Checkpoint 4\n");


	return 0;
}

int mount_fs(char *name, int major)
{
	char *fs_type = NULL;
	char *fs_dir = NULL;
	char *fs_dev = NULL;
	int flags = major;
	int result = 0;

	char *temp = strdup(name);

	fs_dir = strtok(temp, ";");
	fs_dev = strtok(NULL, ";");
	fs_type = strtok(NULL, ";");

	writelog(" Remounting type %s, dir = %s, dev = %s, flags = %d\n",
		 fs_type, fs_dir, fs_dev, flags);

	result = mount(fs_dev, fs_dir, fs_type, flags, NULL);
	writelog(" mount -t %s %s %s -flags %d returned %d\n",
		 fs_type, fs_dev, fs_dir, flags, result);
	return (0);
}


int parse_args(int argc, char *argv[])
{
	int i;
	int err = 0;
	int start = 3;

	if (argc < 3) {
		writelog("Invalid arg count: %d\n", argc);
		err = -EINVAL;
		goto Cleanup;
	}
	//
	// Capture the parameters
	//

	nook_name = strdup(argv[1]);
	writelog("Nook: %s\n", argv[1]);

	switch (*argv[2]) {
		case 'P':
			if (argc < 5) {
				writelog("Invalid arg count: %d\n", argc);
				err = EINVAL;
				goto Cleanup;
			}
			file = argv[3];
			line = argv[4];
			start = 5;
			writelog("Fault: parameter fault in file %s:%s\n", file, line);
			break;
		case 'R':
			if (argc < 5) {
				writelog("Invalid arg count: %d\n", argc);
				err = EINVAL;
				goto Cleanup;
			}
			file = argv[3];
			line = argv[4];
			start = 5;
			writelog("Fault: resource fault in file %s:%s\n", file, line);
			break;
		case 'M':
			if (argc < 5) {
				writelog("Invalid arg count: %d\n", argc);
				err = EINVAL;
				goto Cleanup;
			}
			file = argv[3];
			line = argv[4];
			start = 5;
			writelog("Fault: enomem fault in file %s:%s\n", file, line);
			break;
		case 'S':
			writelog("Fault: stack fault. err=%s\n", argv[3]);
			start = 4;
			break;
		default:
			writelog("Fault: default fault. err=%s\n", argv[3]);
			start = 4;
	}	
	//writelog("%s, %s, %s -> %d\n", argv[2], argv[3], argv[4], start);


	// Capture each interface needing recovery
	//

	printf("argv[4]=%s, argc=%d\n", argv[4], argc);

	for (i = start; i < argc; i++) {
		char itf_type[40], itf_name[40];
		int major;
		char *tmp, *old_tmp;
		driver_t *d;

		d = &drivers[driver_count];

		tmp = strchr(argv[i], ':');
		if (tmp == NULL) {
			break;
		}
		strncpy(itf_type, argv[i], tmp - argv[i]);
		itf_type[tmp - argv[i]] = '\0';
		writelog("Interface type: %s\n", itf_type);

		drivers[driver_count].interface = strdup(itf_type);

		if (strncmp(itf_type, "file", 4) == 0) {
			d->do_1 = NULL;
			d->do_2 = NULL;
			d->do_3 = NULL;
		} else if (strncmp(itf_type, "module", 6) == 0) {
			d->do_1 = unload_module;
			d->do_2 = load_module;
			d->do_3 = NULL;
		} else if (strncmp(itf_type, "network", 7) == 0) {
			d->do_1 = stop_network;
			d->do_2 = start_network;
			d->do_3 = NULL;
		} else if (strncmp(itf_type, "file-system", 11) == 0) {
			//d->do_1 = unmount_fs;
			d->do_1 = NULL;
			d->do_2 = mount_fs;
			d->do_3 = NULL;
		} else if (strncmp(itf_type, "block-device", 12) == 0) {
			//taken care of while recovering bdops
			d->do_1 = NULL;
			d->do_2 = NULL;
			d->do_3 = NULL;
		} else {
			d->do_1 = NULL;
			d->do_2 = NULL;
			d->do_3 = NULL;
		}

		old_tmp = tmp + 1;
		tmp = strchr(old_tmp, ':');
		if (tmp != NULL) {

			strncpy(itf_name, old_tmp, tmp - old_tmp);
			itf_name[tmp - old_tmp] = '\0';
			if (itf_name[0] != '\0') {
				writelog("Interface name: %s\n", itf_name);
				drivers[driver_count].name =
				    strdup(itf_name);
			}
			old_tmp = tmp + 1;
			if (old_tmp[0] != '\0') {
				major = atoi(old_tmp);
				if (major != -1) {
					writelog("Device major: %d\n",
						 major);
					drivers[driver_count].major =
					    major;
				}
			}
		}
		driver_count++;
	}

      printf("driver_count=%d\n", driver_count);
      Cleanup:
	return (err);
}

typedef struct driver_recovery {
	char *name;
	int (*do_recovery) (char *name, int major, int pass);
} driver_recovery_t;


void do_not_recover()
{
	int err, result;
	driver_t *d = NULL;

	writelog("Do not recover driver %s\n", d->name);

	sleep(5);
	result = nooks_delete_nook(nook_name);
	writelog("Deleting nook %s\n", nook_name);
	if (result == -1) {
		err = errno;
	}
	exit(0);
	return;
}

void replace_driver(char *s, char *c)
{
}

int main(int argc, char *argv[])
{
	int err = 0;
	int i;
	time_t t;

	driver_t *d = NULL;
	int do_recovery = 1;

	open_log();

	time(&t);
	writelog("Time: %s\n", ctime(&t));


	printf("Entering parse_args\n");
	err = parse_args(argc, argv);
	if (err < 0) {
		goto Cleanup;
	}
	printf("Exiting parse_args\n");

	err = config_read(1, NULL, "", "/etc/modules.nooks");
	// BUGBUG: o.k. if does not exist
	//  if (err < 0) {
	//    goto Cleanup;
	//  }


	writelog("Unloading nook %s ... ", nook_name);

	printf("Entering unload_nook\n");
	err = nooks_unload_nook(nook_name);
	printf("Exiting unload_nook\n");
	
	appendlog(err == 0 ? "OK\n" : "FAIL\n");

	syslog_fault();

/* FIXME i assume that drivers are already sorted as follows by the kernel:
 * net | fs, file, block, module
 */

/* release ressources */
	printf("Checkpoint 1 %d\n", driver_count);	

	for (i = 0; i < driver_count; i++) {
		d = &drivers[i];

		printf("Checkpoint 2\n");	
		writelog("Releasing driver %s \n",
			 drivers[i].name);

		if (d->do_1) {
			printf("Checkpoint 3\n");	
			d->do_1(d->name, d->major);
		} else {
			writelog(" nothing to do.\n");
		}
	}

	flush_log();

/* load policies for each drivers and deduct whether we recover or not */

	writelog("Loading drivers configuration ... %d drivers\n",
		 driver_count);

	for (i = 0; i < driver_count; i++) {
		d = &drivers[i];

		writestats(d->name, "Fault\n");

		writelog(" * %s\n", d->name);
		load_driver_conf(d);

		execute_driver_cmds(d, ON_FAULT);
		do_recovery &= do_driver_recover(d);
	}

	if (do_recovery) {
		/* restore ressources */
		for (i = (driver_count - 1); i >= 0; i--) {
			d = &drivers[i];

			writelog("Restoring driver %s \n", d->name);

			writestats(d->name, "Recover\n");

			if (d->do_2) {
				d->do_2(d->name, d->major);
			} else {
				writelog(" nothing to do.\n");
			}
			flush_log();
		}

		syslog_recover();


		/* cleanup ressources ? */
		for (i = 0; i < driver_count; i++) {
			d = &drivers[i];
			
			writelog ("Cleaning up driver %s \n", d->name);

			if (d->do_3) {
				d->do_3(d->name, d->major);
			} else {
				writelog(" nothing to do.\n");
			}

			writelog(" execute %s recover_cmds\n", d->name);
			execute_driver_cmds(d, ON_RECOVER);
			flush_log();
		}
	} else {
		writelog(" execute %s last_cmds\n", d->name);
		execute_driver_cmds(d, ON_LAST);
		writelog("Do not recover.\n");
		flush_log();
	}

	writelog("Deleting nook %s ... ", nook_name);

	err = nooks_delete_nook(nook_name);

	appendlog(err == 0 ? "OK\n" : "FAIL\n");

	
	writelog("Job is finished.\n");
	flush_log();
	close_log();


      Cleanup:
	return (err);
}
